from ninja import NinjaAPI
from app_changeset.models import ChangesetModel, FilterModel
from django.http import HttpResponse
from ninja import Schema
from pydantic import Field
from django.shortcuts import get_object_or_404
from typing import List
from typing import Optional
from django.contrib.gis.geos import GEOSGeometry
from django.contrib.gis.geos import Polygon
from decimal import Decimal

import datetime
import json
import osmviews

api = NinjaAPI()

class ChangesetSchema(Schema):
    id: int
    user_id: Optional[int]
    user_name: Optional[str]
    num_changes: Optional[int]
    min_lat: Optional[float]
    max_lat: Optional[float]
    min_lon: Optional[float]
    max_lon: Optional[float]
    created_at: datetime.datetime
    created_at_format_converted: str
    closed_at: str # can be None
    open: bool
    tags: dict
    geom: str 
    status: str
    priority: Optional[float]

    def resolve_created_at_format_converted(self, obj):
        return self.created_at.strftime("%d.%m.%Y,  %H:%M:%S")
    
    def resolve_closed_at(self, obj):
        return str(obj.closed_at)

    def resolve_geom(self, obj):
        return "TODO: proper type conversion: geometry(Polygon,4326) -> some python type"


class FilterAppliedSchema(Schema):
    name: str = None
    user: str = None
    shape: str = None
    tags: str = None
    dateFrom: str = None
    dateTo: str = None
    changesetStatus: str = None

# CRUD filters
class FilterInSchema(Schema):
    name: str = None
    user: str = None
    poly: str = None
    isRectangle: bool = False
    tags: str = None
    dateFrom: str = None
    dateTo: str = None
    changesetStatus: str = None

class FilterOutSchema(Schema):
    id: int
    name: str = None
    user: str = None
    shape: str = None
    isRectangle: bool = False 
    tags: str = None
    dateFrom: datetime.date = None
    dateTo: datetime.date = None
    changesetStatus: str = None

class StatusInSchema(Schema):
    status: str

@api.post("/filters")
def create_filter(request, payload: FilterInSchema):
    # name is not unique

    if(payload.poly is not None):
        payload.poly = GEOSGeometry(payload.poly)

    if(payload.tags is not None):
        payload.tags = json.loads(payload.tags)

    filter = FilterModel.objects.create(**payload.dict())
    return {"name": filter.name}


@api.put("/filters/{filter_name}")
def update_filter(request, filter_name: str, payload: FilterInSchema):

    filter = get_object_or_404(FilterModel, name=filter_name)

    if(payload.poly is not None):
        payload.poly = GEOSGeometry(payload.poly)

    if(payload.tags is not None):
        payload.tags = json.loads(payload.tags)

    for attr, value in payload.dict().items():
        setattr(filter, attr, value)
    filter.save()
    return {"success": True}


@api.delete("/filters/{filter_name}")
def delete_filter(request, filter_name: str):
    filter = get_object_or_404(FilterModel, name=filter_name)
    filter.delete()
    return {"success": True}

@api.get("/filters", response=List[FilterOutSchema])
def list_filters(request):
    tmp = FilterModel.objects.all()

    for i in tmp:
        if i.poly is not None: 
            i.shape = json.dumps(i.poly.geojson)

        if i.tags is not None:
            i.tags = json.dumps(i.tags)

    return tmp


@api.put("/changesets/{changeset_id}")
def update_status(request, changeset_id: int, payload: StatusInSchema):
    changeset = get_object_or_404(ChangesetModel, id=changeset_id)
    changeset.status = payload.status
    changeset.save()
    return {"success": True}
    


@api.post("/changesets", response=List[ChangesetSchema])
def changesets(request, filter: FilterAppliedSchema):

    # filter.user is None or non empty json string
    # filter.shape is None or non empty json Geometry object 
    # filter.tags is None or non empty json string

    query = 'select distinct osm.id, osm.user_id, osm.user_name, osm.num_changes, osm.min_lat, osm.max_lat, osm.min_lon, osm.max_lon, osm.created_at, osm.closed_at, osm.open, osm.tags, osm.status, osm.geom from osm_changeset as osm'
    params = list()


    # JOINs
    if filter.shape is not None:
        query += ' join (' + \
                 'select distinct changeset_id as node_changeset_id from nodes where st_intersects(%s, nodes.geom)' + \
                 ' union distinct' + \
                 ' select distinct extra_nodes.changeset_id as node_changeset_id from nodes join (select distinct unnest(nodes) as id, changeset_id from ways) as extra_nodes on extra_nodes.id = nodes.id where st_intersects(%s, nodes.geom)' + \
                 ') as sub on osm.id = sub.node_changeset_id'


        poly = GEOSGeometry(filter.shape)
        params.append(str(GEOSGeometry(filter.shape)))
        params.append(str(GEOSGeometry(filter.shape)))


    if filter.tags is not None:
        query += ' left join ways on osm.id = ways.changeset_id' + \
                 ' left join nodes on osm.id = nodes.changeset_id' + \
                 ' left join relations on osm.id = relations.changeset_id'

    # WHEREs
    if filter.tags is not None:
        for k,v in json.loads(filter.tags).items():
            query += (' where', ' and')[' where ' in query] + \
                     ' (osm.tags -> %s = %s OR ways.tags ->> %s = %s OR' + \
                     ' nodes.tags ->> %s = %s OR relations.tags ->> %s = %s)'
            params += [k, v, k, v, k, v, k, v]

    if filter.user is not None:
        query += (' where', ' and')[' where ' in query] + ' osm.user_name = %s'
        params.append(filter.user)

    if filter.dateFrom is not None:
        query += (' where', ' and')[' where ' in query] + ' osm.closed_at >= %s'
        params.append(filter.dateFrom)

    if filter.dateTo is not None:
        query += (' where', ' and')[' where ' in query] + ' osm.closed_at <= %s'
        params.append(filter.dateTo)

    if filter.changesetStatus is not None:
        query += (' where', ' and')[' where ' in query] + ' osm.status = %s'
        params.append(filter.changesetStatus)
    else:
        query += (' where', ' and')[' where ' in query] + ' osm.status not like %s'
        params.append('a%') 


    query += ' order by created_at desc limit 100'

    rawQueryset = ChangesetModel.objects.raw(query, params)

    osmviews.download('/tmp/osmviews.tiff')
    with osmviews.open('/tmp/osmviews.tiff') as o:
        for row in rawQueryset:
            if row.min_lat is None or row.max_lat is None or row.min_lon is None or row.max_lon is None:
                row.priority = 0
            else:
        #        row.priority = 1
                 row.priority = round(o.rank(round(float((row.min_lat + row.max_lat) / 2), 6), round(float((row.min_lon + row.max_lon) / 2), 6)), 2)
    return [row for row in rawQueryset]
